
package handler_name;
require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(handler_name dataset_regexes);

use strict;

my $debug = 0;
my $test = 0;

# Lookup the handler name using the regular expressions from the dods.ini
# file. The dods.ini file contains lines with one regular expresion followed
# by the handler name token. Comment start with the `#' character. If the
# string (usually a pathname) matches a regular expression `, this function
# returns the handler associated with that regular expression. If no regexs
# match, then return "".

sub handler_name {
    my ($pathname, $server_config_file) = @_;

    print STDERR "Pathanme: ", $pathname, "\n" if $debug;

    open (DODSINI, $server_config_file);

  LINE: 
    while (<DODSINI>) {
	next LINE if /^\#/;
	next LINE if /^$/;

	my ($regex, $handler) = split;

	print STDERR "Regex: $regex, Handler: $handler \n" if $debug >= 2;
	if ($pathname =~ $regex) {
	    print STDERR "match $handler\n" if $debug >= 1;
	    return $handler;
	}
    }

    return "";
}

# Collect the regexes that describe the files this server is configured to
# recognize as data. Return the regexes in a list. Exclude those that
# are used for non file-based data sources (like JGOFS). 
#
# server_config_file: the name of the DODS server configuration file.
# exclude: a list of handlers whose regexes should be excluded from the
# returned list.

sub dataset_regexes {
    my ($server_config_file, @exclude) = @_;
    my @regexes;			# Used for the return value.

    open (DODSINI, $server_config_file);

  LINE: 
    while (<DODSINI>) {
	next LINE if /^\#/;
	next LINE if /^$/;

	my ($regex, $handler) = split;

	# Note that I'm using $_ explicitly since that contains the pattern I
	# want to test. Usually you don't need to do this. In most cases,
	# $handler is the pattern and @exclude holds the things to be
	# matched. In this case, however, @exclude holds the patterns and I
	# want to see if any of them match the string $handler. The same
	# problem could be solved by concatenating a bunch of patterns with
	# logical ORs, but the resulting pattern could get really large.
	if (!grep {$handler =~ $_} @exclude) {
	    print STDERR "Including $regex\n" if $debug >= 1;
	    @regexes = (@regexes, $regex);
	}
    }

    return @regexes;
}

# Tests
if ($test) {
    ("hdf" eq handler_name("/stuff/file.HDF", "./dods.ini")) || die;
    ("hdf" eq handler_name("/stuff/file.hdf", "./dods.ini")) || die;
    ("nc" eq handler_name("/stuff/file.nc", "./dods.ini")) || die;
    ("nc" eq handler_name("/stuff/file.NC", "./dods.ini")) || die;
    ("nc" eq handler_name("/stuff/file.cdf", "./dods.ini")) || die;
    ("nc" eq handler_name("/stuff/file.CDF", "./dods.ini")) || die;
    ("jg" eq handler_name("/stuff/test", "./dods.ini")) || die;
    ("jg" eq handler_name("/stuff/stuff", "./dods.ini")) || die;
    ("mat" eq handler_name("/stuff/test.mat", "./dods.ini")) || die;
    ("mat" eq handler_name("/stuff/test.Mat", "./dods.ini")) || die;
    ("mat" eq handler_name("/stuff/test.MAT", "./dods.ini")) || die;
    ("ff" eq handler_name("/stuff/test.dat", "./dods.ini")) || die;
    ("ff" eq handler_name("/stuff/test.bin", "./dods.ini")) || die;
    ("" eq handler_name("/stuff/file.bob", "./dods.ini")) || die;
    ("hdf" eq 
     handler_name("/usr/tmp/dods_cache#home#httpd#html#data#hdf#S1700101.HDF",
		  "./dods.ini")) || die;

#      (compare_arrays(\(""),
#       \dataset_regexes("./dods.ini", ("hdf", "nc", "mat",
#  				     "ff", "dsp", "jg")))) || die;

#      (compare_arrays(\('.*\.(HDF|hdf)(.Z|.gz)*\$'), 
#  		    \dataset_regexes("./dods.ini", ("nc", "mat", "ff",
#  						    "dsp", "jg")))) || die;
#      (compare_arrays(\('.*\.(HDF|hdf)(.Z|.gz)*\$' '.*\.(NC|nc|cdf|CDF)$' 
#  		    '.*\.(mat|Mat|MAT)$' '.*\.(dat|bin)$' 
#  		    '.*\.(pvu)(.Z|.gz)*$'),
#  		    \dataset_regexes("./dods.ini", ("jg")))) || die;

#  sub compare_arrays {
#      my ($first, $second) = @_;
#      no warnings;  # silence spurious -w undef complaints
#      return 0 unless @$first == @$second;
#      for (my $i = 0; $i < @$first; $i++) {
#  	return 0 if $first->[$i] ne $second->[$i];
#      }
#      return 1;
#  }
    
    print STDERR "first: ", dataset_regexes("./dods.ini", ("hdf", "nc", "mat",
					       "ff", "dsp", "jg")), "\n";
    print STDERR "second: ", dataset_regexes("./dods.ini", ("nc", "mat", "ff",
					       "dsp", "jg")), "\n";
    print STDERR "third: ", dataset_regexes("./dods.ini", ("jg")), "\n";

    print "All tests successful.\n";
}

1;

# $Log: handler_name.pm,v $
# Revision 1.3.4.3  2001/05/15 23:18:57  jimg
# Added some testing code.
#
# Revision 1.3.4.2  2001/05/09 21:34:34  jimg
# Fixed an error in the EXPORT line.
#
# Revision 1.3.4.1  2001/05/09 21:33:41  jimg
# Added dataset_regexes(...). This returns a list of regular expressions which
# the server is configured to recognize as datasets. There's a way to exclude
# some of the expressions. The routine does not have any automated tests yet,
# but it should...
#
# Revision 1.3  2000/10/19 23:47:09  jimg
# Added more tests and a debug switch.
#
# Revision 1.2  2000/10/18 22:40:17  jimg
# Added the config file as a parameter. This will provide a way for the
# dispatch script to pass in the config file's name.
#
# Revision 1.1  2000/10/18 22:08:01  jimg
# Added
#
