
// (c) COPYRIGHT URI/MIT 1996,1997,1999
// Please read the full copyright statement in the file COPYRIGHT.
//
// Authors:
//      jhrg,jimg       James Gallagher (jgallagher@gso.uri.edu)

// Implementation of the Gui class. A GUI is a program running in a subprocess
// which reads commands from its stdin and writes responses to stdout. The
// GUI presents information to the user using the usual combination of
// graphical candy.
//
// jhrg 6/21/96

#ifdef __GNUG__
#pragma implementation
#endif

#include "config_dap.h"

static char rcsid[] not_used = {"$Id: Gui.cc,v 1.19.4.2 2001/10/08 23:48:09 jimg Exp $"};

#include <stdio.h>

#include <string>

// includes for the Tcl / Tk code
#include <tcl.h>
#include <tk.h>

#include "Gui.h"
#include "util.h"		// For downcase()
#include "escaping.h"
#include "debug.h"
#include "dods_gui.h"

// Initialization function defined in dods_gui.c (built by mktclapp).
// 12/15/99 jhrg
extern "C" int Et_Init(int argc, char **argv);


// dods_gui.c expects to call this after initializing the tcl/tk interpreter.
// This dummy function keeps the linker happy and gives the dods_gui.c code
// something harmless to do. The real functionality of this is embedded in
// the methods of this class. After sending a command to the interpreter, the
// Tcl_DoOneEvent() is called to process all the pending events in the queue.
// 12/15/99 jhrg
extern "C" void
Et_CustomMainLoop(Tcl_Interp *interp)
{
    while (Tcl_DoOneEvent(TCL_ALL_EVENTS | TCL_DONT_WAIT) != 0);
}

Gui::Gui(const string cmd): _show_gui(true), _progress_visible(false)
{
    const char* g = getenv("DODS_USE_GUI");

    string dods_gui = (g ? g : "");
    downcase(dods_gui);
    if (dods_gui == "no")
	_user_wants_gui = false; // Users explicitly say "no"...
    else
	_user_wants_gui = true; // Otherwise assume "yes".

    if (!_user_wants_gui)
	return;

    // now set up the Tcl interpreter
    char *argv[] = {"dods_gui"};
    int argc = 1;
    if (Et_Init(argc, argv) != TCL_OK) {
	_show_gui = false;
        return;
    }

    // execute the initial command if one was given
    if (cmd != "")
	Tcl_Eval(Et_Interp, (char *)(cmd.c_str()));	

    // process all pending events 
    Et_CustomMainLoop(Et_Interp);
}

Gui::Gui(): _show_gui(true), _progress_visible(false)
{
    const char* g = getenv("DODS_USE_GUI");

    string dods_gui = (g ? g : "");
    downcase(dods_gui);
    if (dods_gui == "no")
	_user_wants_gui = false; // Users explicitly say "no"...
    else
	_user_wants_gui = true; // Otherwise assume "yes".

    if (!_user_wants_gui)
	return;

    // now set up the Tcl interpter
    char *argv[] = {{"dods_gui"}};
    int argc = 1;
    if (Et_Init(argc, argv) != TCL_OK) {
	_show_gui = false;
        return;
    }

    // process all pending events 
    Et_CustomMainLoop(Et_Interp);
}

Gui::~Gui()
{
    // Don't try to stop what isn't running
    if (!show_gui() || (Et_Interp == NULL)) 
	return;
    
    _show_gui = false;
    Tcl_DeleteInterp(Et_Interp);
    Et_Interp = NULL;
    return;
}

bool
Gui::progress_visible(int state)
{
    if (state < 0)
	return _progress_visible;
    else
	return (_progress_visible = state);
}

bool
Gui::command(string command)
{
    if (!show_gui())		// Bail if we're not supposed to be here...
	return true;

    DBG(cerr << "Sending the GUI: " << command << endl);

    int rc = Tcl_Eval(Et_Interp, (char *)(command.c_str()));

    // process all pending events 
    Et_CustomMainLoop(Et_Interp);

    // even if Tcl_Eval barfs, still process events
    if (rc != TCL_OK)
	return false;

    return true;
}

bool
Gui::simple_error(string error_msg)
{
    if (!show_gui())		// Bail if were not supposed to be here...
	return true;

    string message = "simple_error ";
    message += munge_error_message(error_msg) + "\n";

    DBG(cerr << "Sending the GUI: " << message << endl);

    int rc = Tcl_Eval(Et_Interp, const_cast<char *>(message.c_str()));

    // process all pending events 
    Et_CustomMainLoop(Et_Interp);

    if (rc != TCL_OK)
	return false;

    return true;
}

// this is a version of command(), but the args are to get the password
bool
Gui::password(string &user, string &pass)
{
    char *pass_var, *user_var;	// char ptrs for Tcl / Tk

    if (!show_gui())		// Bail if were not supposed to be here...
	return true;

    DBG(cerr << "Sending the GUI: password" << endl);

    int rc = Tcl_Eval(Et_Interp, "password");

    // process all pending events 
    Et_CustomMainLoop(Et_Interp);

    // if this was the password command, we need to get the user
    // and password variables now 
    user_var = Tcl_GetVar(Et_Interp, "progress_usr", TCL_GLOBAL_ONLY);
    pass_var = Tcl_GetVar(Et_Interp, "progress_pwd", TCL_GLOBAL_ONLY);
    // usr_var and pass_var are pointers into Tcl memory space, which
    // is quite volatile, so lets copy them out 
    user = (string)user_var;
    pass = (string)pass_var;

    // even if Tcl_Eval barfs, still process events
    if (rc != TCL_OK)
	return false;

    return true;
}

// this is minor variation of command(), used to retrieve the Cancel button
bool
Gui::percent_bar(double pro, int *usr_cancel)
{
    char *cancel_var;	// char ptr for Tcl / Tk
	
    if (!show_gui())		// Bail if were not supposed to be here...
	return true;

    DBG(cerr << "Sending the GUI: bar" << endl);

    string cmd = "bar ";
    cmd += long_to_string((long)pro, 10);
    cmd += " \n";

    int rc = Tcl_Eval(Et_Interp, const_cast<char *>(cmd.c_str()));

    // process all pending events 
    Et_CustomMainLoop(Et_Interp);

    // get the value of the usr var
    cancel_var = Tcl_GetVar(Et_Interp, "usr_cancel", TCL_GLOBAL_ONLY);
    // cancel_var is a pointer into Tcl memory space, which is quite 
    // volatile, and besides, we need an int
    *usr_cancel = atoi(cancel_var);

    // even if Tcl_Eval barfs, still process events
    if (rc != TCL_OK)
	return false;

    return true;
}

bool
Gui::show_gui(int state)
{
    if (state < 0)
	return _show_gui && _user_wants_gui;
    else
	return (_show_gui = (bool)state) && _user_wants_gui;
}

// Don't remove this!! The code in dods_gui.c (built by mktclapp using
// dods_gui.tcl and dods_gui.mta) depends on it. In addition, the awk program
// munge_dods_gui expects to see lines that mktclapp creates because this is
// here. Sorry about the mess... 12/30/99 jhrg

#ifdef TEST

int 
main(int argc, char *argv[])
{
    // Read a command, send it on to the GUI, then read another. exit on
    // blank input.
    string line;
    Gui gui;
    string r1, r2;

    while (1) {
	cout << "gui: ";
	getline(cin, line);
	line += "\n";
	if (line == "\n")
	    exit(0);

	if (line == "password\n") {
	    gui.password(&r1, &r2);
	    cerr << r1 << ", " << r2 << endl;
	}
	else
	    gui.command(line);
    }
}

#endif

// $Log: Gui.cc,v $
// Revision 1.19.4.2  2001/10/08 23:48:09  jimg
// Replaced call to sprintf with calls to string's concatenation operator.
//
// Revision 1.19.4.1  2001/07/28 01:10:42  jimg
// Some of the numeric type classes did not have copy ctors or operator=.
// I added those where they were needed.
// In every place where delete (or delete []) was called, I set the pointer
// just deleted to zero. Thus if for some reason delete is called again
// before new memory is allocated there won't be a mysterious crash. This is
// just good form when using delete.
// I added calls to www2id and id2www where appropriate. The DAP now handles
// making sure that names are escaped and unescaped as needed. Connect is
// set to handle CEs that contain names as they are in the dataset (see the
// comments/Log there). Servers should not handle escaping or unescaping
// characters on their own.
//
// Revision 1.19  2000/09/22 02:17:20  jimg
// Rearranged source files so that the CVS logs appear at the end rather than
// the start. Also made the ifdef guard symbols use the same naming scheme and
// wrapped headers included in other headers in those guard symbols (to cut
// down on extraneous file processing - See Lakos).
//
// Revision 1.18  2000/04/07 00:19:05  jimg
// Merged Brent's changes for the progress gui - he added a cancel button.
// Also repaired the last of the #ifdef Gui bugs so that we can build Gui
// and non-gui versions of the library that use one set of header files.
//
// Revision 1.17.4.1  2000/04/04 05:00:24  brent
// put a Cancel button the Tcl/Tk GUI
//
// Revision 1.17  1999/12/31 01:02:39  jimg
// Spelling fixes...
//
// Revision 1.16  1999/12/31 00:55:10  jimg
// Fixed up the progress indicator
//
// Revision 1.15  1999/08/09 18:27:34  jimg
// Merged changes from Brent for the Gui code (progress indicator)
//
// Revision 1.14.8.2  1999/08/03 04:29:10  brent
// Gui.cc now contains Tcl code, an external Tcl file is not needed
//
// Revision 1.14.8.1  1999/07/29 05:46:17  brent
// call Tcl / GUI directly from Gui.cc, abandon expect, and consolidate Tcl files
//
// Revision 1.14  1999/04/29 02:29:30  jimg
// Merge of no-gnu branch
//
// Revision 1.13.14.2  1999/02/05 09:32:34  jimg
// Fixed __unused__ so that it not longer clashes with Red Hat 5.2 inlined
// math code.
//
// Revision 1.13.14.1  1999/02/02 21:56:59  jimg
// String to string version
//
// Revision 1.13  1997/06/09 20:58:47  jimg
// Added `etc/' to default GUI initialization command location
// (See _start_gui()).
//
// Revision 1.12  1997/04/18 17:26:22  jimg
// Changed include of String.h so that it uses `<>' quotes instead of `""'
// quotes since the latter messed up the dependencies in the Makefile.
//
// Revision 1.11  1997/03/05 08:26:51  jimg
// Fixed bug with _show_gui -vs- _user_wants_gui fields.
//
// Revision 1.10  1997/02/19 04:51:58  jimg
// Fixed bugs with _show_gui and _user_wants_gui.
//
// Revision 1.9  1997/02/10 02:34:40  jimg
// Changed _start_gui() call to dods_root to match new return type.
//
// Revision 1.8  1996/11/25 03:41:28  jimg
// Fixed dods_root so that it now uses the function (rather that a static
// global which won't be initialized when this library is linked using a
// program other than g++).
//
// Revision 1.7  1996/11/22 00:06:23  jimg
// Increased Expect timeout from 2 to 10 seconds.
//
// Revision 1.6  1996/11/18 21:59:32  jimg
// Error messages from command are more verbose - to simplify debugging.
//
// Revision 1.5  1996/11/13 18:58:40  jimg
// Added missing break statements to switch()s.
//
// Revision 1.4  1996/08/13 18:27:46  jimg
// Added not_used in definition of char rcsid[].
//
// Revision 1.3  1996/07/08 23:19:16  jimg
// Added support for the HAVE_EXPECT compile-time switch. When defined as 1
// support for expect is built, when 0 it is not. Thus the dap++ library may be
// built so that expect is not required to link programs. However, the GUI will
// work only when expect/tcl are present and when client are linked with those
// libraries in addition to libdap++.a
// Fixed comments.
// Fixed some warnings (but left some...).
//
// Revision 1.2  1996/06/22 00:04:23  jimg
// Removed forced DEBUG define.
//
// Revision 1.1  1996/06/21 23:12:29  jimg
// Created based on code that was in Connect.cc.
//

