
#ifdef __GNUG__
#pragma implementation
#endif

#include <stdio.h>

#include <iostream>
#include <strstream>

#include "Error.h"
#include "InternalErr.h"
#include "HTTPCache.h"

#define DODS_DEBUG 1
#include "debug.h"

#ifdef Win32
#define MKDIR(a,b)	mkdir((a))
#define REMOVE(a)	remove((a))
#define DIR_SEPARATOR_CHAR	'\\'
#define DIR_SEPARATOR_STR	"\\"
#else
#define MKDIR(a,b)	mkdir((a), (b))
#define REMOVE(a)	remove((a))
#define DIR_SEPARATOR_CHAR	'/'
#define DIR_SEPARATOR_STR	"/"
#endif

/* This is the default cache directory: */
#ifdef Win32
#define HT_CACHE_LOC	"\\tmp\\"
#define HT_CACHE_ROOT	"w3c-cache\\"
#else
#define HT_CACHE_LOC	"/tmp/"
#define HT_CACHE_ROOT	"w3c-cache/"
#endif
#define HT_CACHE_INDEX	".index"
#define HT_CACHE_LOCK	".lock"
#define HT_CACHE_META	".meta"
#define HT_CACHE_EMPTY_ETAG	"@w3c@"

/* Default heuristics cache expirations - thanks to Jeff Mogul for good comments! */
#define NO_LM_EXPIRATION	24*3600		/* 24 hours */
#define MAX_LM_EXPIRATION	48*3600		/* Max expiration from LM */

/*
**  If using LM to find the expiration then take 10% and no more than
**  MAX_LM_EXPIRATION.
*/
#ifndef LM_EXPIRATION
#define LM_EXPIRATION(t)	(HTMIN((MAX_LM_EXPIRATION), (t) / 10))
#endif

#define WARN_HEURISTICS		24*3600		  /* When to issue a warning */

#define DUMP_FREQUENCY	10			 /* Dump index every x loads */

#define MEGA			0x100000L
#define HT_CACHE_TOTAL_SIZE	20		/* Default cache size is 20M */
#define HT_CACHE_FOLDER_PCT	10    /* 10% of cache size for metainfo etc. */
#define HT_CACHE_GC_PCT		10        /* 10% of cache size free after GC */
#define HT_MIN_CACHE_TOTAL_SIZE	 5			/* 5M Min cache size */
#define HT_MAX_CACHE_ENTRY_SIZE	 3     /* 3M Max sixe of single cached entry */

/** Create a HTTP 1.1 compliant cache. This initializes the both the cache
    root and the path to the index file. However, it does not read the index
    file; use the <c>cache_index_read</c> method for that. Check the return
    value of that method to determine if there actually was an index file to
    read. 
    @param cache_root The fully qualified pathname of the directory which
    will hold the cache data. 
    @see cache_index_read */
HTTPCache::HTTPCache(string cache_root) : 
    d_cache_root(cache_root), d_cache_index(cache_root + HT_CACHE_INDEX),
    d_locked_open_file(0)
{
    // Initialize the cache table.
    for (int i = 0; i < CACHE_TABLE_SIZE; ++i)
	d_cache_table[i] = 0;
}

/** Destroy an instance of HTTPCache. This frees the cache table structure. */
HTTPCache::~HTTPCache()
{
    for (int i = 0; i < CACHE_TABLE_SIZE; ++i) {
	if (d_cache_table[i]) {
	    CachePointers *cp = d_cache_table[i];
	    // delete each entry
	    for (CachePointersIter cpi = cp->begin(); cpi != cp->end(); ++cpi)
		delete(*cpi);
	    // now delete the vector that held the entries
	    delete d_cache_table[i];
	}
    }
}

/**@{ @name Cache Index 
    
    These methods manage the cache's index file. Each cache holds an index
    file named <c>.index</c> which stores the cache's state information. */

/** Remove the cache index file. 
    @return True if the file was delete, otherwise false. */
bool
HTTPCache::cache_index_delete()
{
    return (remove(d_cache_index.c_str()) == 0);
}

/** Read the saved set of cached entries from disk. Consistency between the
    in-memory cache and the index is maintained by only reading the index
    file when the HTTPCache object is created. 
    @return True when a cache index was found and read, false otehrwise. */
bool
HTTPCache::cache_index_read()
{
    FILE *fp = fopen(d_cache_index.c_str(), "r");
    // If the cache index can't be opened that's OK; start with an empty
    // cache. 09/05/02 jhrg
    if (!fp)
	return false;

    char line[1024];
    while (!feof(fp) && fgets(line, 1024, fp)) {
	add_entry_to_cache_table(cache_index_parse_line(line));
	DBG2(cerr << line << endl);
    }
    
    return true;
}

/** Add a pointer to the CacheEntry to the cache table. A private method. */
void
HTTPCache::add_entry_to_cache_table(HTTPCache::CacheEntry *entry)
{
    int hash = entry->hash;

    if (!d_cache_table[hash])
	d_cache_table[hash] = new CachePointers;

    d_cache_table[hash]->push_back(entry);
}

/** Get a pointer to a CacheEntry from the cache table. A private method */
HTTPCache::CacheEntry *
HTTPCache::get_entry_from_cache_table(int hash, const string &url)
{
    if (d_cache_table[hash]) {
	CachePointers *cp = d_cache_table[hash];
	for (CachePointersIter i = cp->begin(); i != cp->end(); ++i)
	    if ((*i)->url == url)
		return *i;
    }

    return 0;
}

/** Load one line of index file. A private method. */
HTTPCache::CacheEntry *
HTTPCache::cache_index_parse_line(char *line) throw(InternalErr)
{
	// Read the line and create the cache object
	CacheEntry *entry = new CacheEntry;
	istrstream iss(line);

	iss >> entry->url;
	iss >> entry->cachename;

	iss >> entry->etag;
	if (entry->etag == HT_CACHE_EMPTY_ETAG)
	    entry->etag = "";

	iss >> entry->lm;
	iss >> entry->expires;
	iss >> entry->size;

	iss >> entry->range;

	iss >> entry->hash;
	iss >> entry->hits;
	iss >> entry->freshness_lifetime;
	iss >> entry->response_time;
	iss >> entry->corrected_initial_age;

	iss >> entry->must_revalidate;


	return entry;
#if 0
	// Create the new anchor and fill in the expire information we have
	// read in the index.
	if (cache) {
	    HTAnchor * anchor = HTAnchor_findAddress(cache->url);
	    HTParentAnchor * parent = HTAnchor_parent(anchor);
	    HTAnchor_setExpires(parent, cache->expires);	    
	    HTAnchor_setLastModified(parent, cache->lm);
	    if (cache->etag) HTAnchor_setEtag(parent, cache->etag);
	}
#endif
}

/** Walk through the list of cached objects and save them to disk. The
    current contents of the cache file are written to the index file. If the
    file does not exist, it is created. If the file does exist, we override
    the existing copy but that is OK as we have already read its contents. 

    @exception Error if the index file cannot be opened for writing. */
void
HTTPCache::cache_index_write() throw(Error)
{
    DBG(cerr << "Cache Index. Writing index " << d_cache_index << endl);

    // Open the file for writing. Note - we don't take a backup! This should
    // probably be fixed!
    FILE * fp = NULL;
    if ((fp = fopen(d_cache_index.c_str(), "wb")) == NULL) {
	throw Error(string("Cache Index. Can't open `") + d_cache_index 
		    + string("' for writing"));
    }

    /* Walk through the list and write it out. The format is really
       simple as we keep it all in ASCII. */
    for (int cnt = 0; cnt < CACHE_TABLE_SIZE; cnt++) {
	CachePointers *cp;
	if ((cp = d_cache_table[cnt])) {
	    for (CachePointersIter cpi=cp->begin(); cpi != cp->end(); ++cpi) {
		if (fprintf(fp, 
			    "%s %s %s %ld %ld %ld %c %d %d %ld %ld %ld %c\r\n",
			    (*cpi)->url.c_str(),
			    (*cpi)->cachename.c_str(),
			    (*cpi)->etag == "" ? HT_CACHE_EMPTY_ETAG 
			    : (*cpi)->etag.c_str(),
			    (long) ((*cpi)->lm),
			    (long) ((*cpi)->expires),
			    (*cpi)->size,
			    (*cpi)->range ? '1' : '0',
			    (*cpi)->hash,
			    (*cpi)->hits,
			    (long) ((*cpi)->freshness_lifetime),
			    (long) ((*cpi)->response_time),
			    (long) ((*cpi)->corrected_initial_age),
			    (*cpi)->must_revalidate ? '1' : '0') < 0) {
		    throw Error("Cache Index. Error writing cache index\n");
		}
	    }
	}
    }

    /* Done writing */
    fclose(fp);
}

/** @} End of the cache index documentation */

/** Create the cache's root directory. Paths must always end in
    DIR_SEPARATOR_CHAR. This is a private method.
    @exception Error if the given pathname cannot be created. */
void
HTTPCache::create_cache_root(const string &cache_root) throw(Error)
{
    struct stat stat_info;
    string::size_type cur = 0;

#ifdef Win32
    cur = cache_root[1] == ':' ? 3 : 1;
#else
    cur = 1;
#endif
    while ((cur = cache_root.find(DIR_SEPARATOR_CHAR, cur))) {
	string dir = cache_root.substr(0, cur);
 	if (stat(dir.c_str(), &stat_info) == -1) {
	    DBG(cerr << "Cache....... Creating " << dir << endl);
	    if (MKDIR(dir.c_str(), 0777) < 0) {
		DBG(cerr << "Error: can't create." << endl);
		throw Error(string("Could not create the directory for the cache. Failed when building path at ") + dir + string("."));
	    }
	} else {
	    DBG(cerr << "Cache....... Found " << dir << endl);
	}
	cur++;
    }
}

/** Set the cache's root directory to the given path. If no path is given,
    look at the DODS_CACHE, TMP and TEMP environment variables to guess at a
    good location. If that fails, or if the path cannot be created, throw an
    exception. 

    Note that in most cases callers should look for this path in the user's
    .dodsrc file.

    @see RCReader
    @exception Error if the path can neither be deduced nor created. */
void
HTTPCache::set_cache_root(const string &root = "") throw(Error)
{
    if (root != "") {
	d_cache_root = root;
	// cache root should end in /.
	if (d_cache_root[d_cache_root.size()-1] != DIR_SEPARATOR_CHAR)
	    d_cache_root += DIR_SEPARATOR_CHAR;
    } else {
	// If no cache root has been indicated then look for a suitable
	// location.
	char * cr = (char *) getenv("WWW_CACHE");
	if (!cr) cr = (char *) getenv("TMP");
	if (!cr) cr = (char *) getenv("TEMP");
	if (!cr) cr = HT_CACHE_LOC;

	d_cache_root = cr;
	if (d_cache_root[d_cache_root.size()-1] != DIR_SEPARATOR_CHAR)
	    d_cache_root += DIR_SEPARATOR_CHAR;
	
	d_cache_root += HT_CACHE_ROOT;
    }

    create_cache_root(d_cache_root);
}

/** Get the current cache root directory. */
string
HTTPCache::get_cache_root()
{
    return d_cache_root;
}

/** Lock the cache. Return true if the cache lock was acquired, false
    otherwise. This is a single user cache, so it requires locking at the
    process level. 

    @return True if the cache was locked for our use, False otherwise. */
bool
HTTPCache::get_single_user_lock(bool force) 
{
   if (!d_locked_open_file) {
	FILE * fp;
	string lock = d_cache_root + HT_CACHE_LOCK;
	if ((fp = fopen(lock.c_str(), "r")) != NULL) {
	    fclose(fp);
	    if (force)
		REMOVE(lock.c_str());
	    else
		return false;
	}
	if ((fp = fopen(lock.c_str(), "w")) == NULL) {
	    return false;
	}
	d_locked_open_file = fp;
	return true;
    }

    return false;
}

// $Log: HTTPCache.cc,v $
// Revision 1.1.2.2  2002/09/17 00:18:49  jimg
// Added methods to manage the single user lock and to get/set the cache root
// directory. Also fixed a build problem; I appear to have the sstream headers
// and classes on my machine but other (most?) gcc 2.95.x installations lack
// them. I switched from sstream to strstream.
//
// Revision 1.1.2.1  2002/09/12 22:39:35  jimg
// Initial version.
//
