
// -*- C++ -*-

#include <cppunit/TextTestRunner.h>
#include <cppunit/extensions/TestFactoryRegistry.h>
#include <cppunit/extensions/HelperMacros.h>

#include <string>

#include <Regex.h>

#include "HTTPConnect.h"
#include "RCReader.h"

using namespace CppUnit;

class HTTPConnectTest : public TestFixture {
private:
    HTTPConnect *http;

protected:
    bool re_match(Regex &r, const char *s) {
	return r.match(s, strlen(s)) == (int)strlen(s);
    }

public:
    HTTPConnectTest() {}
    ~HTTPConnectTest() {}
    
    void setUp () {
	http = new HTTPConnect(RCReader::instance());
    }

    void tearDown() {
	delete http;
    }

    CPPUNIT_TEST_SUITE( HTTPConnectTest );

    CPPUNIT_TEST(fetch_url_test);
    CPPUNIT_TEST(get_response_headers_test);
    CPPUNIT_TEST(server_version_test);
    CPPUNIT_TEST(type_test);
    CPPUNIT_TEST(is_reponse_present_test);

    CPPUNIT_TEST_SUITE_END();

    void fetch_url_test() {
	char c;
	try {
	    http->fetch_url("http://dcz.dods.org/");
	    FILE *stuff = http->output();
	    CPPUNIT_ASSERT(fread(&c, 1, 1, stuff) == 1 && !ferror(stuff) && !feof(stuff));

	    http->fetch_url("http://eddy.gso.uri.edu/cgi-bin/nph-dods/avhrr/2001/4/d01093165826.pvu.Z.das");
	    stuff = http->output();
	    CPPUNIT_ASSERT(fread(&c, 1, 1, stuff) == 1 && !ferror(stuff) && !feof(stuff));
	}
	catch (Error &e) {
	    cerr << "Error: " << e.get_error_message() << endl;
	    CPPUNIT_ASSERT(!"Caught a DODS exception from fetch_url");
	}
	catch (InternalErr &e) {
	    cerr << "InternalErr: " << e.get_error_message() << endl;
	    CPPUNIT_ASSERT(!"Caught a DODS exception from fetch_url");
	}
	catch (...) {
	    CPPUNIT_ASSERT(!"Caught an unknown exception from fetch_url");
	}
    }

    void get_response_headers_test() {
	http->fetch_url("http://eddy.gso.uri.edu/cgi-bin/nph-dods/avhrr/2001/4/d01093165826.pvu.Z.das");

	try {
	    vector<string> h = http->get_response_headers();

	    // Should get five headers back.
	    Regex header("XDODS-Server: DAP/.*");
	    CPPUNIT_ASSERT(re_match(header, h[0].c_str()));
	    CPPUNIT_ASSERT(h[4] == "Content-Description: dods_das");
	    CPPUNIT_ASSERT(h.size() == 5);
	}
	catch (InternalErr &e) {
	    CPPUNIT_ASSERT(!"Caught an exception from get_response_headers");
	}
    }

    void server_version_test() {
	http->fetch_url("http://eddy.gso.uri.edu/cgi-bin/nph-dods/avhrr/2001/4/d01093165826.pvu.Z.das");
	Regex version("dap/[0-9]+\\.[0-9]+\\.[0-9]+");
	try {
	    CPPUNIT_ASSERT(re_match(version, http->server_version().c_str()));
	}
	catch (InternalErr &e) {
	    CPPUNIT_ASSERT(!"Caught an exception from server_version");
	}
	
    }

    void type_test() {
	http->fetch_url("http://eddy.gso.uri.edu/cgi-bin/nph-dods/avhrr/2001/4/d01093165826.pvu.Z.das");
	try {
	    CPPUNIT_ASSERT(http->type() == dods_das);
	}
	catch (InternalErr &e) {
	    CPPUNIT_ASSERT(!"Caught an exception from type()");
	}

    }

    void set_credentials_test() {
	http->set_credentials("jimg", "was_quit");
	http->fetch_url("http://dcz.dods.org/secret");
	try {
	    char c;
	    FILE *stuff = http->output();
	    CPPUNIT_ASSERT(fread(&c, 1, 1, stuff) == 1 && !ferror(stuff) && !feof(stuff));
	}
	catch (InternalErr &e) {
	    CPPUNIT_ASSERT(!"Caught exception from output");
	}
    }

    void is_reponse_present_test() {
	http->fetch_url("http://dcz.dods.org/");
	CPPUNIT_ASSERT(http->is_response_present());
	http->close_output();
	CPPUNIT_ASSERT(!http->is_response_present());
	try {
	    http->get_response_headers();
	    CPPUNIT_ASSERT(false && "Didn't throw exception");
	}
	catch (InternalErr &e) {
#if 0
	    cerr << e.get_error_message() << endl;
#endif
	    CPPUNIT_ASSERT(true);
	}
    }
};

CPPUNIT_TEST_SUITE_REGISTRATION(HTTPConnectTest);

int 
main( int argc, char* argv[] )
{
    CppUnit::TextTestRunner runner;
    runner.addTest( CppUnit::TestFactoryRegistry::getRegistry().makeTest() );

    runner.run();

    return 0;
}

// $Log: HTTPConnectTest.cc,v $
// Revision 1.1.2.2  2002/09/14 03:41:31  jimg
// Added tests for get_response_headers and is_response_present.
//
// Revision 1.1.2.1  2002/07/25 13:24:44  jimg
// Added.
//

